// version from 30.03.98
/*================================================================*/
/*                NETWORKS WITH ACTIVE NEURONS                    */
/*================================================================*/
//#include "stdafx.h"

#include "analog.h"
neuron *analog::an_neuron;

analog::analog(int *layers,dataset *d)
{
long i,size,*end;

an_data=d->take(0);

an_active=layers[2];
an_input=layers[0]+layers[2];
an_layer=layers[1];

an_sample=an_data->rows();

an_hidden=an_input;
an_error=an_berr=c_large_value;

an_output=an_input*an_layer; //can be changed (substituted an_hidden)

an_outp=new table(an_sample,an_output);
if(an_output==NULL)memory_error(sizeof(table),"analog::analog"); // check for critical memory points

an_err=new table(an_layer,an_input);

//Dimensions of data
size=2*an_input;
  an_property=new long[size];
  if(an_property==NULL)memory_error(size*sizeof(long),"analog::analog"); // check for critical memory points

  end=an_property;  end+=an_input;
  an_work=end;      end+=an_input;
if(end!=an_property+size)error("It should not happen in analog::analog");


size=an_sample*c_comp_method*an_input;
  an_analogs=new long[size];
  if(an_analogs==NULL)memory_error(size*sizeof(long),"analog::analog"); // check for critical memory points

an_neuron=new neuron(an_data,an_outp,an_err,an_property,an_input);
}

analog::~analog()
{
if(an_outp!=NULL)delete an_outp;
if(an_err!=NULL)delete an_err; //tables
if(an_property!=NULL)delete an_property;
if(an_analogs!=NULL)delete an_analogs;
if(an_neuron!=NULL)delete an_neuron; // neurelement
}

void analog::gl_calculation()
{
long nn,layer;

if(p_optimum){an_neuron->reduse_sorting(); //sorting reduce
             memmove((an_property+an_active),an_property,(an_input-an_active)*sizeof(long));
             for(nn=0;nn<an_active;nn++) an_property[nn]=c_on;}
else
    for(nn=0;nn<an_input;nn++) an_property[nn]=c_on;

data_organize(); // copy input data in table


layer=0;

 do{
    layer++;
    for(nn=0;nn<an_input;nn++){
           if(an_property[nn]==c_on){
              an_neuron->convergence(layer,nn);
              an_neuron->predict(layer,nn,an_analogs);
              }
           }
    an_error=*(an_err->row(layer-1));
    an_berr=layer>1?(*an_err->row(layer-2)):1.0;

    if(p_show)
      fprintf(p_file[0],"\n L:%2d Cur. Error: %8.5f Prev. Error:%8.5f",(int)layer,an_error,an_berr);

    }while(layer<(an_layer-1));
//   }while(an_error<an_berr&&layer<an_layer);

full_predict(layer); //output all result
}


void analog::data_organize() // copy input data in table
{
long i,j,in_row,in_col,ou_row,ou_col;
double *in,*ou,*input,*output,*mark,*end;
table *val;

if(p_print==2)an_data->print(p_file[1]);

val=an_data->values();in_row=val->rows(); in_col=val->columns();
input=val->row(0);

ou_col=an_outp->columns();
output=an_outp->row(0); //beginning

// Copy values in firt column of output table
for(i=0,in=input,ou=output;i<in_row;i++,in+=in_col,ou+=ou_col)
                                     for(j=0;j<in_col;j++) ou[j]=in[j];

an_input=in_col;// count input;

output+=in_col;  //set pointer
in_row=an_data->rows(); in_col=an_data->columns(); input=an_data->row(0);

an_input+=in_col;

// Copy input data in the table
for(i=0,in=input,ou=output;i<in_row;i++,in+=in_col,ou+=ou_col)
                                     for(j=0;j<in_col;j++) ou[j]=in[j];


if(p_print==2){
   for(i=0,ou=an_outp->row(0);i<in_row;i++,ou+=ou_col){
       fprintf(p_file[1],"\n%3d",(int)(i+1));
       for(j=0;j<an_input;j++)fprintf(p_file[1],"%8.5f",ou[j]);
       }
   }
}


void analog::full_predict(long layer) //output all result
{
long i,j;
double *err,*e;

err=an_err->row(0);

fprintf(p_file[1],"\nLAYER");
for(i=0;i<an_input;i++)
           if(an_property[i]==c_on)fprintf(p_file[1],"  PAR:%2d",(int)i);

for(i=0,e=err;i<layer;i++,e+=an_input){
       fprintf(p_file[1],"\n%3d. ",(int)i);
       for(j=0;j<an_input;j++)
           if(an_property[j]==c_on)fprintf(p_file[1],"%8.5f",e[j]);
       }
}

/*====================================================*/
/*    FUNCTIONS FOR CLASS  NEURON                     */
/*====================================================*/
neuron::neuron(data*d,table*ou,table*err,long *prop,long input)
{
long i,size,*end;
double *end1;

n_data=d; n_outp=ou; n_err=err; //set pointers
n_property=prop;

n_inputs=input;
n_active=n_data->values()->columns();
n_output=n_outp->columns();
n_sample=n_data->rows();

size=4*n_output;
  n_priznak=new long[size];
  if(n_priznak==NULL)memory_error(size*sizeof(long),"neuron::neuron"); // check for critical memory points

  end=n_priznak; end+=n_output;
  n_lmodel=end;  end+=n_output;
  n_rmodel=end;  end+=n_output;
  n_work=end;    end+=n_output;
if(end!=n_priznak+size)error("It should not happen in neuron::neuron");

size=3*n_output;

  n_distance=new double[size];
  if(n_distance==NULL)memory_error(size*sizeof(long),"neuron::neuron"); // check for critical memory points

  end1=n_distance; end1+=n_output;
  n_activity=end1; end1+=n_output;
  n_mean_val=end1; end1+=n_output;
  if(end1!=n_distance+size)error("It should not happen in neuron::neuron");

  n_an_space=new double[n_sample];
  if(n_an_space==NULL)memory_error(size*sizeof(long),"neuron::neuron"); // check for critical memory points

n_model=new plong[n_inputs];
if(n_model==NULL)memory_error(size*sizeof(long),"neuron::neuron"); // check for critical memory points

for (i=0;i<n_inputs;i++){ n_model[i]=new long[n_inputs];
       if(n_model[i]==NULL) memory_error(n_inputs*sizeof(long),"dataclaster::dataclaster"); // check for critical memory points
       }
}

neuron::~neuron()
{
int i;
n_outp=n_err=NULL;
delete n_priznak; delete n_distance; delete n_an_space;
for(i=0;i<n_inputs;i++) if(n_model[i])delete n_model[i]; delete n_model;
}

/*====================================================================*/
/* Function: Selection parameters                                     */
/*====================================================================*/
void neuron::convergence(long layer,long xi)
{
long i,k,n,st,nb,inputs,lay;
double delta,dbest,dcrit;
double cur_err,prev_err;

inputs=layer*n_inputs;

for(n=0;n<n_output;n++)n_priznak[n]=c_two; // turn of all parameters

for(i=0;i<n_inputs;i++){
       if(i!=xi)n_priznak[i]=c_off;
       if(n_property[i]==c_on&&i!=xi)
                  for(n=i+n_inputs;n<inputs;n+=n_inputs) n_priznak[n]=c_off;
       }

dcrit=c_large_value;

for(st=0;st<inputs;st++){  dbest=c_large_value; nb=-1;
    for(n=0;n<inputs;n++){
        if(n_priznak[n]==c_off){
           n_priznak[n]=c_on; //include  parameter in analysis
           delta=complexation(inputs,xi);
           if(dbest>delta){dbest=delta; nb=n;memcpy(n_work,n_priznak,n_output*sizeof(long));}
           n_priznak[n]=c_off; // exclude the parameter
           }
       }
   if(nb!=-1)n_priznak[nb]=c_on; //best parameter
   if(dcrit>dbest){dcrit=dbest;memcpy(n_lmodel,n_work,n_output*sizeof(long));}
   }

memcpy(n_priznak,n_lmodel,n_output*sizeof(long));//copy all parameters
}   /* End of function  */
/*====================================================================*/

double neuron::complexation(long inputs,long xi)
{
long i,j,k,na;
double *va,*vi,a,dmin,mse,vv,disp,rr;

mse=disp=0.;
n_mean_val[xi]=n_outp->mean_column(xi);

for(i=0;i<n_sample;i++){
    dmin=c_large_value;  vi=n_outp->row(i);
    for(j=0;j<n_sample;j++){
        va=n_outp->row(j);
        for(k=0,a=0.;k<inputs;k++)
            a+=(n_priznak[k]==c_on)?(vi[k]-va[k])*(vi[k]-va[k]):0.;
        a=sqrt(a);
        if(dmin>a&&i!=j){dmin=a; na=j;}
        }

    vi=n_outp->row_col(i,xi); va=n_outp->row_col(na,xi);

    a=(*vi-*va); vv=(*vi-n_mean_val[xi]);
    disp+=(vv*vv); mse+=a*a;
    }

rr=sqrt(mse/disp);

return rr;
}

void neuron::predict(long layer,long xi,long *t)
{
long i,j,k,row,am_input,*ant;
static long na[c_comp_method],size;

size=n_inputs*c_comp_method;

double *vi,*vm,*err,a,dk_sum,an_sum,mse,vv,disp,rr,model,fs;
static double dmin[c_comp_method],*va[c_comp_method];

am_input=layer*n_inputs;
n_mean_val[xi]=n_outp->mean_column(xi);

dk_sum=an_sum=0.; mse=disp=0.;
ant=t+(xi*c_comp_method);

for(i=0;i<n_sample;i++,ant+=size){
    dmin[0]=c_large_value;  vi=n_outp->row(i);
    n_an_space[i]=c_large_value;

    for(j=0;j<n_sample;j++){
        vm=n_outp->row(j);
        for(k=0,a=0.;k<am_input;k++)
            a+=(n_priznak[k]==c_on)?(vi[k]-vm[k])*(vi[k]-vm[k]):0.;
        a=sqrt(a);
        if(dmin[0]>a&&i!=j){dmin[0]=a; na[0]=j;}
        if(i!=j)n_an_space[j]=a;
        }

     dmin[1]=c_large_value;
     n_an_space[na[0]]=c_large_value;
     for(k=0;k<n_sample;k++)
               if(dmin[1]>n_an_space[k]){dmin[1]=n_an_space[k];na[1]=k;}

      dmin[2]=c_large_value;
      n_an_space[na[1]]=c_large_value;
      for(k=0;k<n_sample;k++)
               if(dmin[2]>n_an_space[k]){dmin[2]=n_an_space[k];na[2]=k;}

       va[0]=n_outp->row_col(na[0],xi);
       va[1]=n_outp->row_col(na[1],xi);
       va[2]=n_outp->row_col(na[2],xi);
       ant[0]=na[0]; ant[1]=na[1]; ant[2]=na[2];


       if(p_method==c_complex_one)model=*va[0];
       else if(p_method==c_complex_two){
               fs=dmin[0]+dmin[1]; fs=fs?fs:1;
               model=dmin[0]*(*va[1])/fs+dmin[1]*(*va[0])/fs;
               }
       else if(p_method==c_complex_three){
               fs=2*(dmin[0]+dmin[1]+dmin[2]); fs=fs?fs:1;
               model=(*va[0]*(dmin[1]+dmin[2])/fs)+
               (*va[1]*(dmin[0]+dmin[2])/fs)+
               (*va[2]*(dmin[0]+dmin[1])/fs);
               }
        else
            error("It is unknown parameter p_method in function neuron::predict()\n",p_method);

        vi=n_outp->row_col(i,xi);
        vm=n_outp->row_col(i,(am_input+xi));

        a=(*vi-model); dk_sum+=(a*a);an_sum+=(*vi)*(*vi);
        vv=(*vi-n_mean_val[xi]); disp+=(vv*vv); mse+=(a*a);
        *vm=model; //record in table
        }

dk_sum/=an_sum; rr=sqrt(mse/disp);
disp=(disp-mse)/disp; mse=sqrt(mse/(n_sample));

err=n_err->row_col(layer-1,xi); *err=rr;

if(p_show)fprintf(p_file[0],"\n L:%2d PARAM:%3d KR=%8.5f  MSE=%8.5f q^2=%8.5f RR=%8.5f ",(int)layer,(int)xi,dk_sum,mse,disp,rr);
if(!xi&&p_print)fprintf(p_file[1],"\n L:%2d KR=%8.5f  MSE=%8.5f q^2=%8.5f RR=%8.5f ",(int)layer,dk_sum,mse,disp,rr);
}


void neuron::reduse_sorting()
{
long n,st,nbest,input;
double d_activ,b_activ,dpred;

input=n_inputs-n_active;

for(n=0;n<n_inputs;n++) n_priznak[n]=0; // set all parameters in zero
dpred=c_large_value;

for(st=0;st<input;st++){
    b_activ=c_large_value; nbest=-1;

    for(n=0;n<input;n++){ n_activity[n]=0.;
        if(n_priznak[n]==c_off){
           n_priznak[n]=c_on; //include  parameter in analysis
           d_activ=complex_one(c_on);

           if(b_activ>d_activ){b_activ=d_activ; nbest=n;
                               memcpy(n_work,n_priznak,n_inputs*sizeof(long));}
           n_activity[n]=d_activ;
           n_priznak[n]=c_off; // exclude the parameter
           }
       }
   n_priznak[nbest]=c_on; //best parameter

   if(p_print) fprintf(p_file[1],"\n PARAM:%3d Error:%8.5f",(int)(nbest+1),n_activity[nbest]);
   if(p_show)  fprintf(p_file[0],"\n PARAM:%3d Error:%8.5f",(int)(nbest+1),n_activity[nbest]);
   if(p_print==2){ fprintf(p_file[1],"  Mask:");for(n=0;n<input;n++)fprintf(p_file[1],"%2d",n_work[n]);}

   if(dpred>b_activ){dpred=b_activ;
                     memcpy(n_lmodel,n_work,n_inputs*sizeof(long));}
   }


memcpy(n_property,n_lmodel,n_inputs*sizeof(long));
memcpy(n_priznak,n_lmodel,n_inputs*sizeof(long));

if(p_print){fprintf(p_file[1]," \n\n Best combination Error:%8.5f ",dpred);
            fprintf(p_file[1],"\nParameters:");
            for(n=0;n<input;n++)if(n_property[n]==c_on)fprintf(p_file[1],"%3d",(int)n+1);
            }

d_activ=complex_one(c_on,1);
}   /* End of function  */
/*====================================================================*/

/*====================================================================*/
/*   Function: Optimization amount of parameters (method analogs)     */
/*====================================================================*/
void neuron::total_sorting()
{
long **sw,i,j,k,step,v,g,m_contr,count,na,type,am_comb,input;
double dk,best_com,dk_best,dk1,dk_best1;

sw=n_model;
FILE *out=p_file[1];
fprintf(out," \n Table of sorting for all parameters from %i ",(int)n_inputs);

//  Analysis of all parameters
for(i=0;i<n_inputs;i++){n_priznak[i]=c_on; n_work[i]=c_on;}

best_com=complex_one(c_on);

for(step=0;step<(n_inputs/2);step++){ //step of sorting
    dk_best=c_large_value;    dk_best1=c_large_value;am_comb=0;
    for(i=0;i<n_inputs;i++)
            for(j=0;j<n_inputs;j++) sw[i][j]=c_off;
    for(i=0;i<n_inputs;i++)sw[i][i]=c_on;
    g=step;v=step;

    fprintf(out," \n \nSorting of %i&&%i parameters from %i ",(int)(step+1),(int)(n_inputs-(step+1)),(int)n_inputs);
    printf(" \n Sorting of %i&&%i parameters from %i ",(int)(step+1),(int)(n_inputs-(step+1)),(int)n_inputs);

    do{
      count=c_off; am_comb++;
      for(i=0;i<n_inputs;i++)n_priznak[i]=c_off;
      for(i=0;i<=step;i++)
          for(j=i;j<n_inputs;j++)
              if(sw[i][j]==c_on){n_priznak[j]=c_on;count++; break;}

      if(count==(step+1)){  //search analogs for all comp.
                dk=complex_one(c_on);
                if(dk_best>dk){dk_best=dk;memcpy(n_lmodel,n_priznak,n_inputs*sizeof(long));}
                if(best_com>dk){ best_com=dk;type=c_on;memcpy(n_work,n_priznak,n_inputs*sizeof(long));}

                dk1=complex_one(c_off);
                if(dk_best1>dk1){dk_best1=dk1;memcpy(n_rmodel,n_priznak,n_inputs*sizeof(long));}
                if(best_com>dk1){ best_com=dk1;type=c_off;memcpy(n_work,n_priznak,n_inputs*sizeof(long));}

                if(p_print==2){fprintf(out,"\n");
                               for(k=0;k<n_inputs;k++)fprintf(out,"%3d",n_priznak[k]);
                               fprintf(out," %8.5f %8.5f \n",dk,dk1);}
                }

   reiter:m_contr=c_off; //set markers for new combination
         if(g<n_inputs-1){sw[v][g]=c_off;g++;sw[v][g]=c_on;}
           else if(v>0){  v--;
                for(j=0;j<n_inputs;j++)
                    if(sw[v][j]==c_on){ m_contr=c_on;
                       sw[v][j]=c_off;
                       if(j<n_inputs-1){sw[v][j+1]=c_on;g=j+1;break;}
                       else goto reiter;
                       }
               if(!m_contr)goto reiter;

               for(i=v+1;i<=step;i++)
                   for(j=0;j<n_inputs;j++)
                       if(sw[i][j]==c_on)sw[i][j]=c_off;
               for(i=v+1;i<=step;i++)
                    if(g<n_inputs-1) sw[i][++g]=c_on;
                    else break;
               v=step;
               }
           else if(v==0&&g>=(n_inputs-1)) v=-1;
    }while(v!=-1);

    printf("COMB -> %10d",(int)am_comb);
    if(p_print){
         fprintf(out," \n Combination best   KR=%8.5f ",best_com);
         fprintf(out," \n Combination first  KR=%8.5f ",dk_best);
         fprintf(out,"\n Parameters:");
         for(k=0;k<n_inputs;k++)if(n_lmodel[k]==c_on)fprintf(out,"%4d",(int)k+1);
         fprintf(out," \n Combination second KR=%8.5f ",dk_best1);
         fprintf(out,"\n Parameters:");
         for(k=0;k<n_inputs;k++)if(n_rmodel[k]==c_off)fprintf(out,"%4d",(int)k+1);
         }

    }

memcpy(n_property,n_work,n_inputs*sizeof(long));

if(p_print){fprintf(p_file[1]," \n\n Best combination Error:%8.5f ",best_com);
            fprintf(p_file[1],"\nParameters:");
            for(k=0;k<n_inputs;k++)if(n_property[k]==type)fprintf(p_file[1],"%3d",(int)k+1);
            }

dk=complex_one(type,1);
}   /* End of function  */
/*====================================================================*/

double neuron::complex_one(long tmask,long print)
{
long i,j,k,row,col,na;
double *va,*vi,a,dk_sum,an_sum,dmin,best_com,mse,vv,disp,rr,mean;
table *val;

val=n_data->values();row=val->rows(); col=val->columns();

vi=val->row(0);
mean=mean_value(vi,vi+col*row,col);

dk_sum=an_sum=0.; mse=disp=0.;

for(i=0;i<n_sample;i++){
    dmin=c_large_value;  vi=n_data->row(i);
    for(j=0;j<n_sample;j++){
        va=n_data->row(j);
        for(k=0,a=0.;k<n_inputs;k++)
            a+=(n_priznak[k]==tmask)?(vi[k]-va[k])*(vi[k]-va[k]):0.;
        a=sqrt(a);
        if(dmin>a&&i!=j){dmin=a; na=j;}
        }

    if(print)fprintf(p_file[1],"\n Sample:%3d - An:%3d  DIST=%8.5f ",(int)(i+1),(int)(na+1),dmin);
    vi=val->row(i); va=val->row(na);

    for(k=0,a=0.;k<col;k++,vi++,va++){
            a=(*vi-*va); dk_sum+=a*a; an_sum+=(*vi)*(*vi);
            vv=(*vi-mean); disp+=vv*vv; mse+=a*a;
            }
    }

dk_sum/=an_sum; rr=sqrt(mse/disp);
disp=(disp-mse)/disp; mse=sqrt(mse/(n_sample*col));

if(print){fprintf(p_file[1],"\n KR=%8.5f  MSE=%8.5f q^2=%8.5f RR=%8.5f ",dk_sum,mse,disp,rr);
          fprintf(p_file[0],"\n KR=%8.5f  MSE=%8.5f q^2=%8.5f RR=%8.5f ",dk_sum,mse,disp,rr);
          }

return rr;
}






