#ifndef _define_data_
#define _define_data_
#include "defines.h"

class bin{
int b_size,b_length,*b_array;

	public:	
//functions:
	bin(int);
	bin(int,FILE*);
	void invert();
	void print(FILE *,int opt=1);
	int all_on(int n=-1); 			// return number of "ON" bits
	void set_bits(int points,long *groups,bin *control,double *probability,double *similarity); // to set some bits "ON"

//inline:
	~bin(){if(b_array!=NULL)delete b_array;}
	int size(){return b_size;} // number of used bytes
	int length(){return b_length;} // number of bits
	void copy(bin*n){//TT if(b_size>n->b_size)error("In bit copy");
					memcpy(b_array,n->b_array,b_size*sizeof(int)); b_length=n->b_length;}
	void on(int i){//TT check_number(i,b_length);
						b_array[i/c_int_bit]|=1<<(i&(c_int_bit-1));}
	void off(int i){//TT check_number(i,b_length);
						b_array[i/c_int_bit]&=~(1<<(i&(c_int_bit-1)));}
	long bit(int i){//TT  check_number(i,b_length);
						return b_array[i/c_int_bit]&(1<<(i&(c_int_bit-1)))?1:0;}
	int	operator[](int i){return bit(i);}
	void clear(){memset(b_array,0,b_size*sizeof(int));}			// to set all elements to OFF
	void set()  {memset(b_array,UCHAR_MAX,b_size*sizeof(int));}	// to set all elements to ON
	void set_on(bin*b);	// set all bins to "ON" if, b[i]=="ON"
	void set_off(bin*b);// inverse operation to set_on()
	int option(){return	all_on()*2<b_length?1:-1;}

};


class table{
long t_rows,t_columns;// can be changed
long t_amount;		// row*column is a maximum alowed value
double	*t_elem,	// elem[row=0,column=0],elem[row=0,column=1]
		*t_norm;	// coeficients for renorm operation
long tt_rows,tt_columns; // parameters to save temporary configuration
double *tt_elem;
	public:

//inline:
	table(long r,long c){t_rows=r;t_columns=c;t_amount=r*c;t_elem=new double[t_amount];t_norm=NULL;
						if(t_elem==NULL)error("not enough memory for %i bytes in table",t_amount);}
	~table(){delete t_elem;if(t_norm!=NULL)delete t_norm;}
	double *renorm(){return t_norm;}
	long columns(){return t_columns;}
	long rows(){return t_rows;}
	void change_settings(long r,long c){if(r*c>t_amount)error("In table::change_setings r=%i c=%i t_amount=%i",r,c,t_amount);t_rows=r;t_columns=c;}
	double *row(long no){//TT check_row(no);
			return t_elem+no*t_columns;} // return pointer on row=no
	double *column(long no){//TT check_col(no);
			return t_elem+no;} // return pointer on column=no
	double *row_col(long nrow,long ncol){TT check_col(ncol);
			return t_elem+nrow*t_columns+ncol;} // return pointer on row =nr +nc
	double elem(long e){//TT if(e>t_columns*t_rows)error("in elem %i>%i",e,t_columns*t_rows);
			return t_elem[e];}
	double *read_diagonal(long no){ //TT check_row(no);
			return t_elem+no*(t_columns+1);} // return pointer on a diagonal element
	void insert_col(double *v,long r){memcpy(row(r),v,sizeof(double)*t_columns);}
	void check_col(long no){if((no&&no>=t_columns)||no<0)error("in table::check_col %i>t_columns=%i",no,t_columns);}
	void check_row(long no){if((no&&no>=t_rows)||no<0)error("in table::check_row %i>t_rows=%i",no,t_rows);}
	double	mmin(long col){return t_elem[no_min(col)*t_columns+col];}
	double	mmax(long col){return t_elem[no_max(col)*t_columns+col];}
	double mean_column(long no){ //TT check_col(no);	// Summa of elements
			return mean_value(t_elem+no,t_elem+t_columns*t_rows,t_columns);}
	double cor_column(long n1,long n2,long gr=1){//TT check_col(n1);check_col(n2); // correlation coefficient between columns i and j
			return correlation(t_elem+n1,t_elem+n2,t_elem+t_columns*t_rows,t_columns,gr);}
	double disp_column(long no,double *mean=(double*)NULL){ //TT check_col(no);		// dispersion
			return dispersion(t_elem+no,t_elem+t_columns*t_rows,(mean==NULL?mean_column(no):*mean),t_columns);}
	void save_cfg()   {tt_rows=t_rows;tt_columns=t_columns;tt_elem=t_elem;}
	void restore_cfg(){t_rows=tt_rows;t_columns=tt_columns;t_elem=tt_elem;}
	void copy(table *t){if(t->t_amount>t_amount)error("In table::copy t_amount=%i > %i",t->t_amount,t_amount);
			memcpy(t_elem,t->t_elem,t->t_amount*sizeof(double));t_rows=t->t_rows;t_columns=t->t_columns;}
			// duplicate the table
//functions:
	void extract_means();
	double *mean_columns();			// return an array of mean values of each column
	double *disp_columns(double *m=(double*)NULL); // return an array of dispersion of each column
	void prune(long n);				//	Remove column=n
	void reduce(table*,bin*,bin*s=(bin*)NULL);//	The new table will contain elements with bin1&&2[i]=0
	void reduce(table*,int*);//	The new table will contain elements with bin1&&2[i]=0
	void print(FILE*,long r=-1,double *rnorm=NULL);		// Print table in file
	long no_max(long col);			// no of maximal element in column col
	long no_min(long col);			// no of minimal element in column col
	void norm_column(long,double,double);	// norm the column
	bin* decorrelate(double c);		// remove columns (one from a pair) with |rij|>c
	void unit();                    // Matris wish 1 diagonal of elements
	void multipnumber(double);      // Multiplication matrix on number
};


class data:
public table
{
// initial data are saved in prototypes
table *d_dat,*d_val,*d_value;
char **d_names;	// names of rows

	public:
//functions:
	data(FILE*,long row,long col,long val, long name);
	~data();
	void print(FILE*,int no=0);	// to print calculated results or data
	long *eu();			// calculate euclidian distance amid points and return an array of pairs of points

//inline:
//	void generate(bin *s1,bin *s2);
//	void swap_lc(){table *t=d_dat[0];d_dat[0]=d_dat[1];d_dat[1]=t;
//						  t=d_val[0];d_val[0]=d_val[1];d_val[1]=t;} // to change set no 0 and 1
	table * take_data(){//TT check_point(no);
					return d_dat;}
	table * take_val(){//TT check_point(no);
					return d_val;}
	double *get_dat(long r=0){//TT check_point(no);
					return d_dat->row(r);}
	double *get_val(long r=0){//TT check_point(no);
					return d_value!=NULL?d_val->row(r):(double*)NULL;}
	table *values(){return d_value;}
	long inputs(int point){//TT if(point!=-1)check_point(point);
			return point==-1?columns():d_dat->columns();}
	long outputs(int point){//TT if(point!=-1)check_point(point);
			return d_value==NULL?0:d_val->columns();}
	long rows_number(long point){//TT if(point!=-1)check_point(point);
			return point==-1?rows():d_dat->rows();}
	char *name(long n){return d_names!=NULL?d_names[n]:(char*)NULL;}
	long name_length();
};

class dataset{
int d_set,
*d_pruned; // contains an order of pruned parameters last element =-1
data *d_data[c_datasets];

	public:
//functions:
	dataset(char*,long col,long val,long name);
	~dataset();
	void norm(long val,long set, double min[2], double max[2]);	// how many datasets should be used for min/max operation 
	void prune(bin &);			// to prune a number of columns 
	void print(FILE*,int no=0);
	int print_pruned(int, FILE**);
	int check_pruned();
	void statistic(FILE*f);
	long max_size();

//inline:
	int columns(){return (*d_data)->columns();}
	int rows(){return (*d_data)->rows();}
	
	void decorrelate(double val){prune(*(*d_data)->decorrelate(val));}
	int  size(){return d_set;}
	data *take(long i){//TT check_no(i);
					return d_data[i];}
	void check_no(int i){if(i<0||i>=d_set)error("in dataset::check_no %i>=%i (d_set)",i,d_set);}
};

#endif

