!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!                                                                             !
!     Name:  COMBI  version 0.8.00                              19.12.2003    !
!                                                                             !
!     Description:  Combinatorial algorithm of the Group Method of Data       !
!                   Handling (GMDH)  (multilayered combinatorial algorithm    !
!                   with possibility of directed sorting of models)           !
!                                                                             !
!     Reference:    http://www.GMDH.net                                       !
!     Author:       Gregory Ivakhnenko        E-mail: Gai@gmdh.net            !
!                                                                             !
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      Program  COMBI
      real
     +X   [allocatable] (:,:),
     +XT  [allocatable] (:,:),  AA  [allocatable] (:,:),
     +RMin[allocatable] (:),    RMax[allocatable] (:)
      real*8
     +B   [allocatable] (:,:),  C   [allocatable] (:,:),
     +SNor[allocatable] (:,:),  SNorB[allocatable] (:,:),
     +CR  [allocatable] (:),    ForVal [allocatable] (:,:),
     +SNorm[allocatable] (:,:), SNorm2[allocatable] (:,:),
     +VN  [allocatable] (:),    VNy [allocatable] (:),  
     +CO  [allocatable] (:)
      integer
     +IAn [allocatable] (:),    SelAn[allocatable] (:,:), 
     +IAnL [allocatable] (:),   IMv  [allocatable] (:)
      integer*2
     +IMq [allocatable] (:),    In  [allocatable] (:),
     +Im ( 600 ),  Iy ( 30 ),   In_Card ( 15 ),
     +iHr,iMin,iSec, i100th,i100th2, iHr2,iMin2,iSec2, iDay,iMon,iYr  ! Time variables

      real*8          Qmax
      integer         LF
      character*12    Name_In
      character*12    Rep_Name, Card_Name, Format ! ,Form
      character*16    Crit  ( 4 )                 ! Criterion name
      character*3     On_Off( 2 )                 ! 'Off','On'
      character*2     FNo                         ! Report file no.
        
      common /LDat/   JPr                         ! Parameters
      common /LMai1/  JRes,    JFile,   JGraph,   JNorm,  JProt,  LFor,
     +                LAn,     LTest,   LSel
      common /LMai2/  JFun,    JS,      JRegul,   ISe,  ICr, JDiv
      common /LSize/  ML, ML0, ML1, ML11, N, LF, MV,MV1,MV2, Qmax    ! Arrays size constants
      automatic                            ! put everything on stack by default
C**********************************************************************
C        Set constants values
C**********************************************************************
      Card_Name   = 'c:\w\_combi'                                    ! Parameters file name

      On_Off( 1 ) = 'Off'                          ! Parameter - 'off' or 'on'
      On_Off( 2 ) = 'On '
      Crit  ( 1 ) = 'Regularity      '             ! Criterion 'name'
!      Crit  ( 2 ) = 'Cross_Validation'
      Crit  ( 2 ) = 'Unbiased        '
      Crit  ( 3 ) = 'Correlation     '
      Qmax  = 1.E+300                              ! Big constant
      write (*,1111)                               ! Write name board
C**********************************************************************
C        Data and parameters input
C**********************************************************************
      ic = 1
      call CardY_1 ( ic, Card_Name, Name_In, M,N ) ! Card input

      allocate ( In( N ), stat = MemErr )          ! Set In(N) as declared in
      if( MemErr .ne. 0 )  stop '  Abort: Not enough storage!'
   
!      MQ = M                                       
      call CardY_2 ( ic, Name_In, Rep_Name, Format, M,MY,N,
     +               MM,NN,Im,Iy,In, In_Card, 15 ) ! Card input
C-----------------------------------------------------------------------------
      ICr    = In_Card (  1 )  ! main criterion
      LFor   = In_Card (  2 )  ! prediction mode: number of delayed arguments
      Lan    = In_Card (  3 )  ! number of ansambles selected at each step
      LSel   = In_Card (  4 )  ! number of variables to select
      ISe    = In_Card (  5 )  ! selection layer
      JNorm  = In_Card (  6 )  ! normilize variables
      JDiv   = In_Card (  7 )  ! division of observations
      JFun   = In_Card (  8 )  ! skip usage of functions as variables
      JS     = In_Card (  9 )  ! cancel procedure, when criterion is increasing
      LTest  = In_Card ( 10 )  ! number of hidden points for validation of model
      JRegul = In_Card ( 11 )  ! regularization mode no.
!      JGraph = In_Card ( 12 )  ! substract linear trend
      JFile  = In_Card ( 13 )  ! create results file
      JRes   = In_Card ( 14 )  ! output of the additinal results
      JPr    = In_Card ( 15 )  ! write initial data
C-----------------------------------------------------------------------------
      MM0 = MM                        ! Save vars initial size for delayed args creation
      M0  = MM + MY              ! Save initial columns number in sample
     
      if( LFor .ne. 0 ) then
      NT = LTest + 1          ! Validation set C
      if(JFun == 0) then 
         MM = (MM+MY) * LFor     ! Use all output variables
      else
         MM =  MM * LFor         ! New array size with delayed var, delayed y - variable
      endif
   else
      NT = LTest 
   endif
      
      ML = MM + MY                    ! Total array size
      NN = NN - LTest
   
      if( NT == 0 ) NT = 1 
      allocate ( X( ML,NN ), XT( M0,NT ), stat = MemErr )   ! Data array allocation
      if( MemErr.ne.0 ) stop 'Abort: Not enough storage for X( ML,NN )!'
C**********************************************************************
C           Initial array input from file
C**********************************************************************
      call DatYPIn( Name_In, Format, MM0,M0,M,MY,N,NT,ML,MM,NN,X,XT,
     +              Im,In,Iy )                  ! M=MM, N=NN !!!X
      write (*,332)  M
      write (*,333) MY
      write (*,334)  N
C**********************************************************************
C           Set arrays bounds
C**********************************************************************
      if( LFor .ne. 0 .or. JFun .ne. 0 )  then                  
         ML11 = MM               ! Variables limit column no.
      else
         ML11 = ML - 1           ! For approximation
      endif

      ML0 = ML                !!! Function column no.
      ML1 = ML11 + 2             ! allocate number for AA,SNOR...

      M1 = M + 1
      if(LAn == 0) LAn = M * 25     ! Set no. of selected models
      LF = LAn                        ! Number of CR_min saved in CR,SelAn

      if( ISe == 0 .and. M > 5 .and. LSel .ne. 0 ) ISe = 5 ! Set selection layer
      if( ISe == 0 ) LSel = 0
      if( LSel > ML11 ) LSel = ML11
C**********************************************************************
C           Write messages
C**********************************************************************
      if( JPr .eq. 0 ) then            !<-"Write initial data - On/Off"
         write (*,11)
      else
         write (*,22)
      endif
      write (*,111) On_Off( JFile+1 ),  ! On_Off( JGraph+1 ), 
     +      On_Off( JRes+1 ), On_Off( JNorm +1 ), 
     +      On_Off( 2-JFun ), On_Off( 2-JS )

      if( LFor .ne. 0 ) then
         write (*,33)  LFor
      else
         write (*,55)
      endif

      if( ICr .gt. 2 )  ICr = 1                 ! Check CR type
      write (*,999)  Crit( ICr+1 )

      if( JRegul .eq. 0 ) then
         write (*,61)                                ! MAPE criterion
      elseif( JRegul .eq. 1 ) then
         write (*,62)                                ! Variation criterion
      elseif( JRegul .eq. 2 ) then
         write (*,63)                                ! Unbiased criterion
      elseif( JRegul .eq. 3 ) then
         write (*,64)                                ! AR_C criterion
      if( LTest < 3 ) then
   write(*,*)'Abort: For AR_C criterion must be validation sample !'
         stop
      endif          
      elseif( JRegul .eq. 4 ) then
         write (*,65)                                ! Disable regularization
   else
      JRegul = 0
      write (*,61)                                ! else MAPE criterion
      endif

      if(LTest. ne. 0) write (*,888)  LTest
      if( N .le. 12  .or.  M0 .gt. N ) write (*,8888)
      if( JPr .ne. 0 ) pause '*  Press Enter to continue  *'

      if( N < 4 ) stop '*** Abort: Too few points (N < 4)! ***' 
      if( ML11 .le. 1 )
     +            stop '*** Abort: Too few variables (M = 1)! ***' 
C**********************************************************************
C           Memory allocation
C**********************************************************************
      allocate(
     +   AA  ( ML1,N ), SNor( ML1,ML1 ), B ( ML1,ML1 ), C ( ML1,ML1 ),
     +   CO  ( ML1 ),   CR  ( LF ),      RMin( ML ),   RMax ( ML ),
     +   VNy ( N ),     VN  ( N ),       IAn ( ML11 ),  IMq ( ML11 ),    
     +   IMv ( ML11 ),  SelAn (LF,ML11), IAnL ( ML11 ),  
     +   SNorm(ML1,ML1),SNorm2(ML1,ML1), SNorB(ML1,ML1),ForVal( 2,NT ), 
     +   stat=MemErr)

      if( MemErr .ne. 0 )  stop 'Abort: Not enough storage !'
C**********************************************************************
C           Open report file and write to it
C**********************************************************************
      io = 3
      if( JFile .ne. 0 ) then
         J = 0
         goto 200
C-----------------------------------------------------------------------------
9999        J = J + 1                           ! Try to change report name
            if( J .gt. 99 ) stop '* Abort: Cannot open report file ! *'
            write ( FNo,'(I2.2)' )  J
            I = len_trim( Rep_Name )
            Rep_Name = Rep_Name( 1:I-2 ) // FNo
C-----------------------------------------------------------------------------
200      open (io, file= Rep_Name, status='NEW', mode='WRITE', err=9999)
         call GetDat ( iYr, iMon, iDay )
         call GetTim ( iHr, iMin, iSec, i100th )
         write ( io,444, err=9991 ) iHr,iMin,iSec, iDay,iMon,iYr

         write ( io,222 )
         write ( io,331 ) Name_In
         write ( io,332 ) MM0
         write ( io,333 ) MY
         write ( io,334 ) N

         if( LFor .ne. 0 )  then
            write ( io,33 )  LFor
         else
            write ( io,55 )
         endif

         write (io,999)  Crit( ICr+1 )

      if( JRegul .eq. 0 ) then
         write (io,61)                                ! MAPE criterion
      elseif( JRegul .eq. 1 ) then
         write (io,62)                                ! Variation criterion
      elseif( JRegul .eq. 2 ) then
         write (io,63)                                ! Unbiased criterion
      elseif( JRegul .eq. 3 ) then
         write (io,64)                                ! AR_C criterion
      elseif( JRegul .eq. 4 ) then
         write (io,65)                                ! Disable regularization
      endif

         if( LTest. ne. 0) write ( io,888 )  LTest
         if( N .le. 10 .or. M0 .gt. N ) write (io,8888)
         if( JPr .ne. 0 ) then
            write (io,4444)
            call MatPriAF ( io, M0,N, ML,N,600, X, Im )
         endif 
      endif
C**********************************************************************
C           Write data array to screen
C**********************************************************************
300   if( JPr .ne. 0 .and. LFor == 0 ) then        ! Write data array without delayed args
         write (*,4444)
         call MatPriA ( M0,N, ML,N,600, X, Im )
         if( N .le. 30 ) pause '*  Press Enter to continue  *'
         write (*,2222)
         call Vi2Pri ( M0, 600, Im )
      endif
C**********************************************************************
C           Create delayed variables for prediction
C**********************************************************************
      if( LFor .ne. 0 ) then
         call DelayArg ( MM0,ML,M,MY,N, X, Im,RMin )  ! Create delayed variables

         if( JPr .ne. 0 )  then
            if( JFile .ne. 0 )  then               ! Write to file
               write (io,5555) LFor
               call MatPriAF ( io, ML,N, ML,N,600, X, Im )
            endif
            write (*,5555) LFor
            call MatPriA ( ML,N, ML,N,600, X, Im )

            if( N .le. 30 ) pause '*  Press Enter to continue  *'
         endif
      else
      endif
C**********************************************************************
C           Normalization
C**********************************************************************
      if( JNorm .ne. 0 ) then
         call NorMiMax ( ML, N, ML,NN, X, RMax,RMin ) ! Normilize initial data
      call NorMiMa2 ( NT, M0,ML,NT,XT,RMax,RMin )  ! Normilize validation data according to Rmin, Rmax

      if( JFile .ne. 0 ) then
        write (io,76) ( Im( I ),   I=1,ML )     
        write (io,77) ( Rmax( I ), I=1,ML )
        write (io,78) ( Rmin( I ), I=1,ML )
        write (io,79) ( Rmax( I ) - RMin( I ), I= 1,ML )
        write (io,6666)    
      endif

      if( JPr .ge. 2  .and.  N .le. 30 )  then
            write (*,7777)
            call MatPriA ( ML,N, ML,N,600, X, Im )
            pause '*  Press Enter to continue  *'
      else
            write (*,6666)
      endif
      
   elseif( JFile .ne. 0 ) then
      write (io,2222)
      write (io,3333)  ( Im( I ), I= 1,ML )
   endif
C**********************************************************************
C           Call main procedure  MODEL
C**********************************************************************

      call MODEL ( io, M0,M1,MY,NT,X,XT,AA,SNor,SNorB,SNorm,SNorm2,B,C,   !? MQ,NT  ->
     +             CO, CR, VN,VNy,ForVal, RMin,RMax, Im,In,IAnL,
     +             IAn,SelAn,IMq,IMv )
C**********************************************************************
C           Memory deallocation and end of the program
C**********************************************************************
10    deallocate ( X,XT, AA,SNor,B,C, CO,CR, RMin,RMax, VN,VNy,ForVal,
     +    IAnL,IAn,SelAn,IMq,IMv,SNorm,SNorm2,SNorB, stat = MemErr)
      if( MemErr .ne. 0 )  write (*,*) ' Warning: Deallocation error!'

      if( JFile  .ne. 0 ) then
         call GetTim ( iHr2, iMin2, iSec2, i100th2 )

         if( i100th2 .lt. i100th ) then
            i100th2 = i100th2 + 100
            iSec2   = iSec2 - 1
         endif
         if( iSec2 .lt. iSec ) then
            iSec2 = iSec2 + 60
            iMin2 = iMin2 - 1
         endif
         if( iMin2 .lt. iMin ) then
            iMin2 = iMin2 + 60
            iHr2  = iHr2  - 1
         endif
         if( iHr2 .lt. iHr )  iHr2 = iHr2 + 24
         i100th = i100th2 - i100th
         iSec   = iSec2 - iSec
         iMin   = iMin2 - iMin
         iHr    = iHr2  - iHr
         write ( io,44 )  iHr, iMin, iSec, i100th
         write ( io,555 )
         close ( io )
         write ( *,666 ) Rep_Name
      else
         write ( *,555 )
      endif
      stop
C**********************************************************************
C           End of the program
C**********************************************************************
111   format (
!     + 15X, 'Substract linear trend                  -  ', A3,/,
     + 15X, 'Create  report  file              - ', A3,/,
     + 15X, 'Write additional values to report - ', A3,/,
     + 15X, 'Normalize  initial  data on scale - ', A3,/,
     + 15X, 'Use  functions  as  variables     - ', A3,/,
     + 15X, 'Continue, when criterion increase - ', A3  )
11    format(/,10X, 'Parameters :',/,
     +         15X, 'Write  initial  data', 14X, '- Off')
22    format(/,10X, 'Parameters :',/,
     +         15X, 'Write  initial  data', 14X, '- On ')
33    format(/,10X, 'Mode             - Prediction  ( with',I2,
     +              ' delayed arguments )')
55    format(/,10X, 'Mode             - Approximation')
999   format ( 10X, 'Main criterion   - ', A16 )
61    format ( 10X, 'Second criterion - MAPE',/)
62    format ( 10X, 'Second criterion - Variation',/)
63    format ( 10X, 'Second criterion - Unbiased',/)
64    format ( 10X, 'Second criterion - AR_C',/)
65    format ( 10X, 'Second criterion - disabled',/)
888   format ( 10X, 'Number of hidden points for validation =',I3,/)
331   format (/,10X,'Data  file  name -     ',    A12 )
332   format ( 10X, 'Number of variables    M  =', I4 )
333   format ( 10X, 'Number of functions    MY =', I4 )
334   format ( 10X, 'Number of observations N  =', I4 )
6666  format (/,10X,'*  Variables was normalized on scale [-1;1] *')
76 format(/,1X,'Data range for each variable:',/,60(''),/,5X,600I12)
77 format(  1X, 'Max:  ', 600F12.4 )
78 format(  1X, 'Min:  ', 600F12.4 )
79 format(  1X, 'Range:', 600F12.4 )
7777  format(//,10X,'Normalized data sample:'  )
2222  format(//,10X,'Ensemble of the variables and functions used:')
3333  format(   10X, 60(''), /,7X, 600I5,/)
4444  format(/, 10X,'Initial data sample  ',
     +              '(without validation subsample):')
88    format(/, 10X,'Order of ranged observations in the data sample:')
5555  format(/, 10X,'Initial data array with ',I2,' delayed variables:')
8888  format (/, 6X,'*** Small number of points can be a reason of ',
     +              'small model accuracy ! ***',/)
222   format ( 14('*'),
     +  '  Results  of  the  GMDH  Combinatorial  Algorithm  ', 14('*'))
444   format (/,59X, I2,':',I2.2,':',I2.2,2X,I2,'\',I2.2,'\',I4) ! Date
44    format (/, 5X, '*  Elapsed computation time  - ',
     +           I2, ':',I2.2,  ':',I2.2,'s',I2.2, '  *' )
66    format (  10X, '*  Press Enter to continue  *',\ )
99    format (/,10X, '*      Wait a bit, please     *',\ )
555   format ( /,5X, '*  Happy End  *')
666   format ( /,5X, '*  Results are in the report file  -  ', A12,' *')
777   format (/,19X, 'For help on usage of switches type  COMBI ?', /)
1111  format (/,'  ',
     +'ͻ',
     +/,'  ',5X,'Combinatorial Algorithm of the Group Method ',
     +    'of Data Handling (GMDH)',3X,'',
     +/,'  ',12X,'Version 0.8.00    Gregory Ivakhnenko (c),'
     +    '   1989-1995', 10X,'',
     +/,'  ',15X,'Gai@gmdh.net      http://www.GMDH.net',23X,'',
     +/,'  ',
     +'ͼ')

9991  write (*,9992)
9992  format(/,' * Warning: No space on disk for writing in the report',
     +         ' file ! *',/,
     +         ' *',10X,'Results will be not written in it.',16X,'*', /)
      end

!----------------------------------------------------------------------
!     Calculate deviation from linear trend
!----------------------------------------------------------------------
!      AA (1,:) = 1.
!      VNy(:)   = X( -Im( ML0 ),: )             ! Save VNy - function
!  
!  do I= 1,N-NT
!     AA ( 2,I ) = I                       ! Forming of AA array  
!         AA ( 3,I ) = X( -Im( ML0 ),I+NT )
!  enddo
!
!  do I= 1,NT
!     AA ( 2,N-NT+I ) = N-NT+I               ! Forming of AA array   
!         AA ( 3,N-NT+I ) = XT( -Im( ML0 ),I )
!  enddo
!
!      call REG8 ( Kl, 2,3,N, ML1, B,C,SNor,AA, CO )! Found linear trend coeficients
!  
!  Co1 = CO( 1 )
!  Co2 = CO( 2 )
!  Sre = 0.
!      Sr = 0.                            ! Calculate average of function
!     do 70 I= 1,N
!70      Sr = Sr + VNy( I )
!  Sr = Sr / N
!
!  do I= 1,N
!     Sre = Sre + ( VNy( I ) - Sr )** 2      ! Quadrat of deviation from average     
!  enddo 
!      
!  if( Kl .eq. 0 ) then                ! if not solved
!     write (*,99) ( CO( I ), I= 1,2 )        
!     if( JFile .ne. 0 ) write (io,99) ( CO( I ), I= 1,2 )
!
!     SS  = 0.
!          Srq = 0.
!     do 250 I= 1,N
!        SS = CO( 1 ) + I * CO( 2 )                   ! Calculate model value
!250        Srq = Srq + (( X( -Im( ML0 ), I ) - SS )** 2 )  ! Quadrat of deviation from trend
!
!     if( JGraph /= 0 ) then                 ! if needed to substract trend
!        do 20 I= 1,N
!20         X( -Im( ML0 ),I) = X( -Im( ML0 ),I) - CO(1) - I * CO(2)
!
!        write (*,88)
!        if( JFile .ne. 0 ) write (io,88)
!     endif
!  else
!     Srq = Sre
!     endif                                                